// byte, 8-bit unsigned int, undefined, 8-bit byte
function unsignedByte(view, {count, value, offset, littleEndian}) {
  if (count == 1) return view.getUint8(offset + 8, littleEndian);
  let o = count > 4 ? value : (offset + 8);
  let vals = [];
  for (let n=0;n<count;n++) {
    vals[n] = view.getUint8(o + n);
  }
  return vals;
}

function ASCII(view, {count, value, offset}) {
  let start = count > 4 ? value : (offset + 8);
  let str = '';
  for (let n = start; n < start+count-1; n++) {
    str += String.fromCharCode(view.getUint8(n));
  }
  return str;
}

// short 16 bit int
function unsignedShort(view, {count, value, offset, littleEndian}) {
	if (count == 1) return view.getUint16(offset + 8, littleEndian);
	let o = count > 2 ? value : (offset + 8);
	let vals = [];
	for (let n=0; n < count; n++) {
		vals[n] = view.getUint16(o + 2*n, littleEndian);
	}
	return vals;
}

// long, 32 bit int
function unsignedLong(view, {count, value, offset, littleEndian}) {
	if (count == 1) return view.getUint32(offset + 8, littleEndian);
	let vals = [];
	for (let n=0; n < count; n++) {
		vals[n] = view.getUint32(value + 4*n, littleEndian);
	}
	return vals;
}

// unisgned rational
function unsignedRational(view, {count, value, littleEndian}) {
  if (count == 1) {
    let numerator = view.getUint32(value, littleEndian);
    let denominator = view.getUint32(value+4, littleEndian);
    let val = new Number(numerator / denominator);
    val.numerator = numerator;
    val.denominator = denominator;
    return val;
  } else {
    let vals = [];
    for (let n=0; n < count; n++) {
      let numerator = view.getUint32(value + 8*n, littleEndian);
      let denominator = view.getUint32(value+4 + 8*n, littleEndian);
      vals[n] = new Number(numerator / denominator);
      vals[n].numerator = numerator;
      vals[n].denominator = denominator;
    }
    return vals;
  }
}

// long, 32 bit int
function signedLong(view, {count, value, offset, littleEndian}) {
  if (count == 1) return file.getInt32(offset + 8, littleEndian);
  let vals = [];
  for (let n = 0; n < count; n++) {
    vals[n] = view.getInt32(value + 4*n, littleEndian);
  }
  return vals;
}

// long, 32 bit int
function signedRational(view, {count, value, littleEndian}) {
  if (count == 1) return view.getInt32(value, littleEndian) / view.getInt32(value+4, littleEndian);
  let vals = [];
  for (let n=0; n < count; n++) {
    vals[n] = view.getInt32(value + 8*n, littleEndian) / view.getInt32(value+4 + 8*n, littleEndian);
  }
  return vals;
}

const exif = {
	0x8827: 'ISO',
	0x010F: 'Make',
	0x0110: 'Model',
	0x9209: 'Flash',
	0xA500: 'Gamma',
	0x829D: 'FNumber',
	0x8825: 'GPSInfo',
	0x0131: 'Software',
	0xA301: 'SceneType',
	0x0132: 'ModifyDate',
	0x8769: 'ExifOffset',
	0x9290: 'SubSecTime',
	0xA001: 'ColorSpace',
	0x0112: 'Orientation',
	0x011A: 'XResolution',
	0x011B: 'YResolution',
	0x9207: 'MeteringMode',
	0x9202: 'ApertureValue',
	0x9217: 'SensingMethod',
	0x0128: 'ResolutionUnit',
	0x8822: 'ExposureProgram',
	0x9203: 'BrightnessValue',
	0xA002: 'PixelXDimension',
	0xA003: 'PixelYDimension',
	0x9003: 'DateTimeOriginal',
	0x9201: 'ShutterSpeedValue',
	0x9204: 'ExposureBiasValue',
};

const gps = {
	0x0002: 'GPSLatitude',
	0x0006: 'GPSAltitude',
	0x0004: 'GPSLongitude'
};

const orderLittleEndian = 0x4949;

const endianAssertion = 0x002a;

const jpeg = 0xffd8;

const exifMarker = 0xffe1;

const exifId =  0x45786966; // "E", "X", "I", "F"

const firstMarker = 2;

const segment = {
	marker: 0,
	length: 2,
	exifId: 4,
};

const tiffHeader = {
	fromSegment: 10,
	byteOrder: 0,
	endianAssertion: 2,
	ifdOffset: 4,
};

const ifd = {
	fromTiffHeader: -1,
	tag: 0,
	type: 2,
	count: 4,
	value: 8,
};

/**
 * If the input is not JPEG file with Exif information,
 * it returns `undefined`.
 * @see http://www.cipa.jp/std/documents/j/DC-008-2012_J.pdf
 */
async function exifer(input, {parse = {}, exif: exif$1 = {}, gps: gps$1 = {}} = {}) {
	exif$1 = {...exif$1, ...exif};
	gps$1 = {...gps$1, ...gps};

  const view = await prepareDataView(input);

	if (!isValidJpeg(view)) throw new Error(`Invalid JPEG`);

  const segmentOffset = await findExifSegmentOffset(view);
	if (segmentOffset < 0) throw new Error(`Segment not found`);

  const tiffHeaderOffset = segmentOffset + tiffHeader.fromSegment;
  const littleEndian = isLittleEndian(view, tiffHeaderOffset);
  const ifdPosition = findIfdPosition(view, tiffHeaderOffset, littleEndian);

  let ifd0 = readTags(view, ifdPosition, tiffHeaderOffset, littleEndian, exif$1);
  const entries = view.getUint16(ifdPosition, littleEndian);
  const thumbnailPosition = view.getUint32(ifdPosition + 2 + entries * 12, littleEndian);

  if (ifdPosition !== 0)
    ifd0 = {...ifd0, ...readTags(view, thumbnailPosition + 6, tiffHeaderOffset, littleEndian, exif$1)};

  if (ifd0.ExifOffset)
    ifd0 = {...ifd0, ...readTags(view, ifd0.ExifOffset + 6, tiffHeaderOffset, littleEndian, exif$1)};

  if (ifd0.InteropOffset)
    ifd0 = {...ifd0, ...readTags(view, tiffHeaderOffset + ifd0.InteropOffset, tiffHeaderOffset, littleEndian, exif$1)};

  if (ifd0.GPSInfo)
    ifd0 = {...ifd0, ...readTags(view, tiffHeaderOffset + ifd0.GPSInfo, tiffHeaderOffset, littleEndian, gps$1)};

	for (const [key, parser] of Object.entries(parse)) {
		const value = ifd0[key];
		if (value) ifd0[key] = parser(value);
	}

  return ifd0;
}

function readTags(view, ifdPosition, tiffHeaderOffset, littleEndian, tags) {
  const result = {};
  const fieldIterator = iterateIfdFields(view, ifdPosition, littleEndian);
  for (const offset of fieldIterator) {
    const tag = view.getUint16(offset, littleEndian);
    if (!tags[tag]) continue;
    let value = readTagValue(view, offset, tiffHeaderOffset, littleEndian);
    if (value) result[tags[tag]] = value;
  }
  return result;
}

function readTagValue(view, offset, tiffHeaderOffset, littleEndian) {
  const type = view.getUint16(offset + ifd.type, littleEndian);
  const count = view.getUint32(offset + ifd.count, littleEndian);
  const value = view.getUint32(offset + ifd.value, littleEndian) + tiffHeaderOffset;
  const state = { value, offset, littleEndian, count };

  if (type == 1 || type == 7)
    return unsignedByte(view, state);

  if (type == 2)
    return ASCII(view, state);

	if (type == 3)
    return unsignedShort(view, state);

	if (type == 4)
    return unsignedLong(view, state);

	if (type == 5)
    return unsignedRational(view, state);

	if (type == 9)
    return signedLong(view, state);

	if (type == 10)
    return signedRational(view, state);
}

async function prepareDataView(input) {
  let arrayBuffer;
  if (typeof File !== `undefined` && input instanceof File) {
    arrayBuffer = await readFile(input);
  } else if (typeof Buffer !== `undefined` && input instanceof Buffer) {
    arrayBuffer = input.buffer;
  } else {
    arrayBuffer = input;
  }
  return new DataView(arrayBuffer);
}

async function readFile(file) {
  const arrayBuffer = await new Promise((resolve) => {
    const reader = new FileReader();
    reader.onload = () => resolve(reader.result);
    reader.readAsArrayBuffer(file);
  });
  return arrayBuffer;
}

function isValidJpeg(view) {
  return view.byteLength >= 2 && view.getUint16(0, false) === jpeg;
}

async function findExifSegmentOffset (view) {
  for await (const segmentPosition of iterateMarkerSegments(view)) {
    if (isExifSegment(view, segmentPosition)) {
      assertExifSegment(view, segmentPosition);
      return segmentPosition;
    }
  }
  return -1;
}

/**
 * APPx/Exif p.18, 19, 150
 * - marker (short) `0xffe1` = APP1
 * - length (short) of segment
 * - padding (short) `0x0000` if exif
 * - "EXIF" (char[4]) if exif
 * - content
 * (The doc describe APP1 have to lay next to the SOI,
 *  however, Photoshop renders a JPEG file that SOI is followed by APP0.)
 */
async function* iterateMarkerSegments (view) {
  let segmentPosition = firstMarker;
  while (true) {
    yield segmentPosition;
    const offsetLength = segment.length;
    const length = offsetLength + view.getUint16(segmentPosition + offsetLength, false);
    segmentPosition += length;
    if (segmentPosition > view.byteLength) return -1;
  }
}

function isExifSegment (view, segmentPosition) {
  const marker = view.getUint16(segmentPosition + segment.marker, false);
  return marker === exifMarker;
}

function assertExifSegment (view, segmentPosition) {
  // p 150
  const id = view.getUint32(segmentPosition + segment.exifId, false);
  if (id !== exifId) {
    throw new Error(`Segment marked as Exif does not have Exif identifier`);
  }
}

function isLittleEndian(view, tiffHeaderOffset) {
  const endian = view.getUint16(tiffHeaderOffset + tiffHeader.byteOrder, false);
  const littleEndian = endian === orderLittleEndian;
  return littleEndian;
}

/*
 * TIFF Header p.17
 * - byte order (short). `0x4949` = little, `0x4d4d` = big
 * - 42 (0x002a) (short)
 * - offset of IFD (long). Minimum is `0x00000008` (8).
 */
function findIfdPosition(view, tiffHeaderOffset, littleEndian) {
  const endianAssertionValue = view.getUint16(tiffHeaderOffset + tiffHeader.endianAssertion, littleEndian);
  if (endianAssertionValue !== endianAssertion) {
    throw new Error(`Invalid JPEG format: littleEndian ${littleEndian}, assertion: 0x${endianAssertionValue}`);
  }
  const ifdDistance = view.getUint32(tiffHeaderOffset + tiffHeader.ifdOffset, littleEndian);
  return tiffHeaderOffset + ifdDistance;
}

/**
 * IFD p.23
 * - num of IFD fields (short)
 * - IFD:
 *   - tag (short)
 *   - type (short)
 *   - count (long)
 *   - value offset (long)
 * - IFD...
 */
function* iterateIfdFields (view, ifdFieldOffset, littleEndian) {
  const numOfIfdFields = view.getUint16(ifdFieldOffset, littleEndian);
  const fieldLength = 12;
  for (let i = 0; i < numOfIfdFields; i++) {
    const currentOffset = ifdFieldOffset + i * fieldLength + 2;
    yield currentOffset;
  }
}

export default exifer;
